#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

from octobot_backtesting.constants import BACKTESTING_DATA_FILE_SEPARATOR
from octobot_backtesting.collectors.data_collector import DataCollector
from octobot_backtesting.importers.data_importer import DataImporter
from octobot_commons.tentacles_management.class_inspector import get_deep_class_from_parent_subclasses


async def create_importer_from_backtesting_file_name(config, backtesting_file) -> DataImporter:
    collector_klass = get_deep_class_from_parent_subclasses(
        parse_class_name_from_backtesting_file(backtesting_file), DataCollector)
    importer = collector_klass.IMPORTER(config, backtesting_file) if collector_klass else None

    if not importer:
        return None

    await importer.initialize()
    return importer


def parse_class_name_from_backtesting_file(backtesting_file):
    return os.path.basename(backtesting_file).split(BACKTESTING_DATA_FILE_SEPARATOR)[0]



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_backtesting.collectors.exchanges.abstract_exchange_history_collector import \
    AbstractExchangeHistoryCollector
from octobot_commons.tentacles_management.advanced_manager import get_single_deepest_child_class


async def collect_exchange_historical_data(exchange_name, symbols, time_frames=None) -> str:
    collector_class = get_single_deepest_child_class(AbstractExchangeHistoryCollector)
    collector_instance = collector_class({}, exchange_name, symbols, time_frames,
                                         use_all_available_timeframes=time_frames is None)
    await collector_instance.initialize()
    await collector_instance.start()
    return collector_instance.file_name



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_backtesting.api.importer import get_data_timestamp_interval
from octobot_commons.constants import CONFIG_ENABLED_OPTION, MINUTE_TO_SECONDS
from octobot_backtesting.backtesting import Backtesting
from octobot_backtesting.constants import CONFIG_BACKTESTING, CONFIG_BACKTESTING_DATA_FILES
from octobot_commons.enums import TimeFramesMinutes
from octobot_commons.logging.logging_util import get_logger
from octobot_commons.time_frame_manager import find_min_time_frame, get_config_time_frame


async def initialize_backtesting(config, exchange_ids, matrix_id, data_files) -> Backtesting:
    backtesting_instance = Backtesting(config=config,
                                       exchange_ids=exchange_ids,
                                       matrix_id=matrix_id,
                                       backtesting_files=data_files)
    await backtesting_instance.create_importers()
    await backtesting_instance.initialize()

    if not backtesting_instance.importers:
        raise ValueError("No importers created")

    return backtesting_instance


async def initialize_independent_backtesting_config(independent_backtesting) -> dict:
    return await independent_backtesting.initialize_config()


async def modify_backtesting_timestamps(backtesting, set_timestamp=None,
                                        minimum_timestamp=None, maximum_timestamp=None) -> None:
    await backtesting.time_updater.modify(set_timestamp=set_timestamp,
                                          minimum_timestamp=minimum_timestamp,
                                          maximum_timestamp=maximum_timestamp)


async def adapt_backtesting_channels(backtesting, config, importer_class):
    # set mininmum and maximum timestamp according to all importers data
    min_time_frame_to_consider = find_min_time_frame(get_config_time_frame(config))
    importers = backtesting.get_importers(importer_class)
    timestamps = [await get_data_timestamp_interval(importer, min_time_frame_to_consider)
                  for importer in importers]  # [(min, max) ... ]

    await modify_backtesting_timestamps(
        backtesting,
        minimum_timestamp=min(timestamps)[0],
        maximum_timestamp=max(timestamps)[1])
    try:
        from octobot_trading.api.exchange import has_only_ohlcv

        if has_only_ohlcv(importers):
            set_time_updater_interval(backtesting,
                                      TimeFramesMinutes[min_time_frame_to_consider] * MINUTE_TO_SECONDS)
    except ImportError:
        get_logger("BacktestingAPI").error("requires OctoBot-Trading package installed")


def set_time_updater_interval(backtesting, interval_in_seconds):
    backtesting.time_manager.time_interval = interval_in_seconds


async def start_backtesting(backtesting) -> None:
    await backtesting.start_time_updater()


async def stop_backtesting(backtesting) -> None:
    await backtesting.stop()


async def stop_independent_backtesting(independent_backtesting) -> None:
    await independent_backtesting.stop()


def get_importers(backtesting) -> list:
    return backtesting.importers


def get_backtesting_current_time(backtesting) -> float:
    return backtesting.time_manager.current_timestamp


def is_backtesting_enabled(config) -> bool:
    return CONFIG_BACKTESTING in config and CONFIG_ENABLED_OPTION in config[CONFIG_BACKTESTING] \
           and config[CONFIG_BACKTESTING][CONFIG_ENABLED_OPTION]


def get_backtesting_data_files(config) -> list:
    if CONFIG_BACKTESTING in config and CONFIG_BACKTESTING_DATA_FILES in config[CONFIG_BACKTESTING]:
        return config[CONFIG_BACKTESTING][CONFIG_BACKTESTING_DATA_FILES]
    return []



# cython: language_level=3
#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class ChannelsManager:
    cdef object logger

    cdef list exchange_ids
    cdef list producers

    cdef str matrix_id

    cdef int refresh_timeout

    cdef list _get_trading_producers(self)
    cdef list _get_evaluator_producers(self)

cdef bint _check_producers_consumers_emptiness(list producers, int priority_level)
cdef list _get_backtesting_producers()
cdef list _get_channel_producers(object channel)



#  Drakkar-Software OctoBot-Backtesting
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
# from distutils.extension import Extension
import os

from setuptools import dist
dist.Distribution().fetch_build_eggs(['Cython>=0.29.17'])

try:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
except ImportError:
    # create closure for deferred import
    def cythonize(*args, **kwargs):
        from Cython.Build import cythonize
        return cythonize(*args, **kwargs)

    def build_ext(*args, **kwargs):
        from Cython.Distutils import build_ext
        return build_ext(*args, **kwargs)

from setuptools import find_packages
from setuptools import setup, Extension

from octobot_backtesting import PROJECT_NAME, VERSION

PACKAGES = find_packages(exclude=["tests"])

packages_list = ["octobot_backtesting.backtesting",
                 "octobot_backtesting.collectors.data_collector",
                 "octobot_backtesting.collectors.exchanges.exchange_collector",
                 "octobot_backtesting.collectors.social.social_collector",
                 "octobot_backtesting.data.data_file_manager",
                 "octobot_backtesting.data.database",
                 "octobot_backtesting.channels_manager.channels_manager",
                 "octobot_backtesting.channels.time",
                 "octobot_backtesting.data_manager.time_manager",
                 "octobot_backtesting.producers.time_updater",
                 "octobot_backtesting.importers.data_importer",
                 "octobot_backtesting.importers.exchanges.exchange_importer",
                 "octobot_backtesting.importers.social.social_importer",
                 "octobot_backtesting.converters.data_converter",
                 "octobot_backtesting.util.backtesting_util"]

ext_modules = [
    Extension(package, [f"{package.replace('.', '/')}.py"])
    for package in packages_list]

# long description from README file
# with open('README.md', encoding='utf-8') as f:
#     DESCRIPTION = f.read()

REQUIRED = open('requirements.txt').readlines()
REQUIRES_PYTHON = '>=3.7'
CYTHON_DEBUG = False if not os.getenv('CYTHON_DEBUG') else os.getenv('CYTHON_DEBUG')

setup(
    name=PROJECT_NAME,
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Backtesting',
    license='LGPL-3.0',
    author='Drakkar-Software',
    author_email='drakkar-software@protonmail.com',
    description='OctoBot project backtesting engine',
    packages=PACKAGES,
    include_package_data=True,
    # long_description=DESCRIPTION,
    cmdclass={'build_ext': build_ext},
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    data_files=[],
    setup_requires=REQUIRED if not CYTHON_DEBUG else [],
    install_requires=REQUIRED,
    ext_modules=cythonize(ext_modules, gdb_debug=CYTHON_DEBUG),
    python_requires=REQUIRES_PYTHON,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Cython',
    ],
)
